//! Halton sampling

const PRIMES: [u32; 1000] = include!(concat!(env!("OUT_DIR"), "/primes.rs"));
const MAX_HALTON_RESOLUTION: u32 = 128;

pub struct Sampler {
    x_scale: u8,
    x_exponent: u8,
    x_inverse: u8,
    y_scale: u8,
    y_exponent: u8,
    y_inverse: u8,
    index: u32,
    dimension: u16,
}

impl Sampler {
    pub fn new(width: u32, height: u32) -> Self {
        let mut x_scale = 1u8;
        let mut x_exponent = 0u8;
        let x_res = u32::min(width, MAX_HALTON_RESOLUTION) as u8;
        while x_scale < x_res {
            x_scale *= 2;
            x_exponent += 1;
        }

        let mut y_scale = 1u8;
        let mut y_exponent = 0u8;
        let y_res = u32::min(height, MAX_HALTON_RESOLUTION) as u8;
        while y_scale < y_res {
            y_scale *= 3;
            y_exponent += 1;
        }

        let x_inverse = mod_inverse(y_scale, x_scale);
        let y_inverse = mod_inverse(x_scale, y_scale);

        Self {
            dimension: 2,
            index: 0,
            x_scale,
            x_exponent,
            x_inverse,
            y_scale,
            y_exponent,
            y_inverse,
        }
    }

    pub fn start_pixel(&mut self, x: u32, y: u32) {
        let mut index = 0;
        let sample_stride = self.y_scale as u32 * self.x_scale as u32;

        if sample_stride > 1 {
            let mx = x % MAX_HALTON_RESOLUTION;
            let my = y % MAX_HALTON_RESOLUTION;

            let x_offset = inverse_radical_inverse(mx, 2, self.x_exponent as u32);
            let y_offset = inverse_radical_inverse(my, 3, self.y_exponent as u32);

            index += x_offset * (sample_stride / self.x_scale as u32) * self.x_inverse as u32;
            index += y_offset * (sample_stride / self.y_scale as u32) * self.y_inverse as u32;

            index %= sample_stride;
        }

        self.index = index;
    }

    pub fn next_sample(&mut self) {
        self.index += self.x_scale as u32 * self.y_scale as u32;
    }

    pub fn get_pixel_2d(&self) -> (f32, f32) {
        (
            radical_inverse(0, self.index >> self.x_exponent),
            radical_inverse(1, self.index / self.y_scale as u32),
        )
    }

    pub fn get_1d(&mut self) -> f32 {
        if self.dimension as usize >= PRIMES.len() {
            self.dimension = 2;
        }
        let dim = self.dimension as u64;
        self.dimension += 1;
        scrambled_radical_inverse(dim as usize, self.index, mix_bits(1 + (dim << 4)))
    }

    pub fn get_2d(&mut self) -> (f32, f32) {
        if self.dimension as usize + 1 >= PRIMES.len() {
            self.dimension = 2;
        }
        let mut dim = self.dimension as u64;
        self.dimension += 2;
        let x = scrambled_radical_inverse(dim as usize, self.index, mix_bits(1 + (dim << 4)));
        dim += 1;
        let y = scrambled_radical_inverse(dim as usize, self.index, mix_bits(1 + (dim << 4)));
        (x, y)
    }
}

fn egcd(a: i16, b: i16) -> (i16, i16) {
    if a == 0 {
        (0, 1)
    } else {
        let (x, y) = egcd(b % a, a);
        (y - (b / a) * x, x)
    }
}

fn mod_inverse(a: u8, m: u8) -> u8 {
    let m = m as i16;
    let (x, _) = egcd(a as i16, m);
    ((x % m + m) % m) as u8
}

fn radical_inverse(base_index: usize, mut i: u32) -> f32 {
    let base = PRIMES[base_index];
    let inv_base = 1.0 / base as f32;
    let mut inv_base_m = 1.0;
    let mut reversed = 0;

    while i > 0 {
        let next = i / base;
        let digit = i - next * base;
        reversed = reversed * base + digit;
        inv_base_m *= inv_base;
        i = next
    }

    f32::min(reversed as f32 * inv_base_m, 1.0 - f32::EPSILON)
}

fn inverse_radical_inverse(mut inverse: u32, base: u32, digits: u32) -> u32 {
    let mut i = 0;
    for _ in 0..digits {
        let digit = inverse % base;
        inverse /= base;
        i = i * base + digit;
    }
    i
}

fn scrambled_radical_inverse(base_index: usize, mut i: u32, hash: u64) -> f32 {
    let base = PRIMES[base_index];
    let inv_base = 1.0 / base as f32;
    let mut inv_base_m = 1.0;
    let mut reversed = 0;

    while 1.0 - inv_base_m < 1.0 {
        let next = i / base;
        let digit_value = i - next * base;
        let digit_hash = mix_bits(hash ^ digit_value as u64);
        let digit = permutation_element(digit_value, base, digit_hash as u32);
        reversed = reversed * base + digit;
        inv_base_m *= inv_base;
        i = next;
    }

    f32::min(reversed as f32 * inv_base_m, 1.0 - f32::EPSILON)
}

// magic. Taken from pbrt/util/hash.h in pbrt
#[inline(always)]
fn mix_bits(mut val: u64) -> u64 {
    val ^= val >> 31;
    val *= 0x7fb5d329728ea185;
    val ^= val >> 27;
    val *= 0x81dadef4bc2dd44d;
    val ^ (val >> 33)
}

// more magic. Taken from pbrt/util/math.h in pbrt
#[inline(always)]
fn permutation_element(mut i: u32, n: u32, seed: u32) -> u32 {
    let mut w = n - 1;
    w |= w >> 1;
    w |= w >> 2;
    w |= w >> 4;
    w |= w >> 8;
    w |= w >> 16;
    loop {
        i ^= seed;
        i *= 0xe170893d;
        i ^= seed >> 16;
        i ^= (i & w) >> 4;
        i ^= seed >> 8;
        i *= 0x0929eb3f;
        i ^= seed >> 23;
        i ^= (i & w) >> 1;
        i *= 1 | seed >> 27;
        i *= 0x6935fa69;
        i ^= (i & w) >> 11;
        i *= 0x74dcb303;
        i ^= (i & w) >> 2;
        i *= 0x9e501cc3;
        i ^= (i & w) >> 2;
        i *= 0xc860a3df;
        i &= w;
        i ^= i >> 5;
        if i < n {
            break;
        }
    }
    (i + seed) % n
}
